/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * receive.c
 * Functions to handle receiving datagrams
 * Copyright (C) 2006 Simon Newton
 */

#include "private.h"


/*
 * Checks if the callback is defined, if so call it passing the packet and
 * the user supplied data.
 * If the callbacks return a non-zero result, further processing is canceled.
 */
int check_callback(node n, pathport_packet_pdu *p, callback_t callback) {
  if (callback.fh != NULL)
    return callback.fh(n,p, callback.data);

  return PATHPORT_EOK;
}


/*
 * Handle a pathport dmx packet
 */
static int pathport_rx_handle_dmx(node n, pathport_packet_pdu *pdu) {
  pathport_pdu_data *data;
  int len, s, uni;

  // run callback if defined
  if (check_callback(n, pdu, n->callbacks[PATHPORT_DMX_HANDLER]) )
    return PATHPORT_EOK;

  if (n->dmx_callback.fh == NULL)
    return 0;

  data = &pdu->d.data;

  if (data->type == PATHPORT_XDMX_DATA_FLAT) {
    len = ntohs(data->num);
    s = ntohs(data->offset);
    uni = s / PATHPORT_DMX_LENGTH;

//    printf("len %i , offset %i , uni %i, %p\n", len, s, uni, n->dmx[uni]);

    if ( n->dmx[uni] != NULL) {
      memcpy(n->dmx[uni], data->data, len);

      if (n->dmx_callback.fh != NULL)
        n->dmx_callback.fh(n, uni, len, n->dmx[uni], n->dmx_callback.data);

    }

  }
  return 0;
}


/*
 * Handle a pathport get request
 */
static int pathport_rx_handle_get(node n, pathport_ipacket p, pathport_packet_pdu *pdu) {
  uint16_t len = ntohs(pdu->head.len);

//  printf("got get for %i params from %x \n", len/2, p->p.head.src );

//  for (i=0; i<len/2; i++) {
//    printf("param %hx\n" , pdu->d.get.params[i]);
//  }

  return pathport_tx_send_getreply(n , ntohl(p->p.head.src), pdu->d.get.params, len/2 );
//  return pathport_tx_send_getreply(n ,PATHPORT_ID_MCAST_MANAGE  , pdu->d.get.params, len/2 );
}


/*
 * Handles an arp request packet
 */
static int pathport_rx_handle_arpreq(node n, pathport_ipacket p, pathport_packet_pdu *pdu) {

  // run callback if defined
  if (check_callback(n, pdu, n->callbacks[PATHPORT_ARPREQ_HANDLER]) )
    return PATHPORT_EOK;

  pathport_tx_send_arprep(n, PATHPORT_ID_MCAST_MANAGE);

  return PATHPORT_EOK;
}


/*
 * Handle an arp reply packet
 */
static int pathport_rx_handle_arprep(node n, pathport_packet_pdu *pdu) {

  // run callback if defined
  if (check_callback(n, pdu, n->callbacks[PATHPORT_ARPREP_HANDLER]) )
    return PATHPORT_EOK;

  return PATHPORT_EOK;
}


/*
 * Handle a PDU
 *
 * @param n the node
 * @param pdu the pdu
 */
static int pathport_rc_handle_pdu(node n, pathport_ipacket p, pathport_packet_pdu *pdu) {

  uint16_t op = ntohs(pdu->head.type);
  uint16_t len = ntohs(pdu->head.len);

  if (check_callback(n, pdu, n->callbacks[PATHPORT_RECV_HANDLER]) )
    return 0;

  if ( op == PATHPORT_DATA && len >= sizeof(pathport_pdu_data) ) {
    pathport_rx_handle_dmx(n,pdu);
  } else if ( op == PATHPORT_GET) {
    pathport_rx_handle_get(n, p, pdu);
  } else if ( op == PATHPORT_ARPREQ && len == sizeof(pathport_pdu_arpreq) ) {
    pathport_rx_handle_arpreq(n, p, pdu);
  } else if ( op == PATHPORT_ARPREP && len >= sizeof(pathport_pdu_arprep) ) {
    pathport_rx_handle_arprep(n,pdu);
  } else {
    printf("unknown or unimplemented op code %hx\n", op );
  }

  return 0;
}


/*
 * Check a pathport packet for validity
 *
 * @return non-zero if this isn't for us
 */
static int pathport_rx_packet_ok(node n, pathport_packet *p) {

  uint32_t dst = ntohl(p->head.dst);

  int to_us = ( dst == n->state.pathport_id ||
          dst == PATHPORT_ID_BROADCAST ||
          dst == PATHPORT_ID_MCAST_ALL ||
          dst == PATHPORT_ID_MCAST_DATA ||
          dst == PATHPORT_ID_MCAST_MANAGE ) ? 1 : 0;

//  printf("proto %x %x , ver %x %x , us %i \n" ,  ntohs(p->head.proto) , PATHPORT_PROTO ,
//      ntohs(p->head.ver), PATHPORT_VERSION, to_us);

  // validity checks
  if ( ntohs(p->head.proto) != PATHPORT_PROTO ||
    ntohs(p->head.ver) != PATHPORT_VERSION || ! to_us) {
    return FALSE;
  }
  return TRUE;
}


/*
 * the main handler for an pathport packet calls
 * the appropriate handler function
 *
 */
int pathport_rx_handle(node n, pathport_ipacket p) {
  int pdu_size, offset, dlen;
  pathport_packet_pdu *pdu = NULL;

  if ( p->length < sizeof(pathport_packet_header) )
    return PATHPORT_EOK;

  if ( ! pathport_rx_packet_ok(n, &p->p) ) {
//    printf("packet not for me %x \n", ntohl(p->p.head.dst) );
    return 0;
  }

  offset=0;

  // length of all pdus in this packet
  dlen = p->length - sizeof(pathport_packet_header);

  while( offset + sizeof(pathport_pdu_header) <= dlen ) {
    pdu = (pathport_packet_pdu*) &p->p.data[offset];

    pdu_size = ntohs(pdu->head.len);

    if ( offset + sizeof(pathport_pdu_header) + pdu_size <= dlen) {
      pathport_rc_handle_pdu(n, p, pdu);
    }

    offset += pdu_size + sizeof(pathport_pdu_header);
  }
  return PATHPORT_EOK;
}
